@@ -8,16 +8,28 @@ import numpy.typing as npt
 from code_loader.contract.datasetclasses import CustomCallableInterfaceMultiArgs, \
     CustomMultipleReturnCallableInterfaceMultiArgs, ConfusionMatrixCallableInterfaceMultiArgs, CustomCallableInterface, \
     VisualizerCallableInterface, MetadataSectionCallableInterface, PreprocessResponse, SectionCallableInterface, \
-    ConfusionMatrixElement, SamplePreprocessResponse
+    ConfusionMatrixElement, SamplePreprocessResponse, InstanceCallableInterface, ElementInstance
 from code_loader.contract.enums import MetricDirection, LeapDataType, DatasetMetadataType
 from code_loader import leap_binder
+from code_loader.contract.mapping import NodeMapping, NodeMappingType, NodeConnection
 from code_loader.contract.visualizer_classes import LeapImage, LeapImageMask, LeapTextMask, LeapText, LeapGraph, \
     LeapHorizontalBar, LeapImageWithBBox, LeapImageWithHeatmap
 
 
+def _add_mapping_connections(connects_to, arg_names, node_mapping_type, name):
+    for user_unique_name, connection_destinations in connects_to.items():
+        main_node_mapping = NodeMapping(name, node_mapping_type, user_unique_name, arg_names=arg_names)
+        node_inputs = {}
+        for arg_name, destination in zip(arg_names, connection_destinations):
+            node_inputs[arg_name] = destination.node_mapping
+
+        leap_binder.mapping_connections.append(NodeConnection(main_node_mapping, node_inputs))
+
+
 def tensorleap_custom_metric(name: str,
                              direction: Union[MetricDirection, Dict[str, MetricDirection]] = MetricDirection.Downward,
-                             compute_insights: Optional[Union[bool, Dict[str, bool]]] = None):
+                             compute_insights: Optional[Union[bool, Dict[str, bool]]] = None,
+                             connects_to=None):
     def decorating_function(user_function: Union[CustomCallableInterfaceMultiArgs,
     CustomMultipleReturnCallableInterfaceMultiArgs,
     ConfusionMatrixCallableInterfaceMultiArgs]):
@@ -28,6 +40,10 @@ def tensorleap_custom_metric(name: str,
 
         leap_binder.add_custom_metric(user_function, name, direction, compute_insights)
 
+        if connects_to is not None:
+            arg_names = leap_binder.setup_container.metrics[-1].metric_handler_data.arg_names
+            _add_mapping_connections(connects_to, arg_names, NodeMappingType.Metric, name)
+
         def _validate_input_args(*args, **kwargs) -> None:
             for i, arg in enumerate(args):
                 assert isinstance(arg, (np.ndarray, SamplePreprocessResponse)), (
@@ -116,7 +132,8 @@ def tensorleap_custom_metric(name: str,
 
 
 def tensorleap_custom_visualizer(name: str, visualizer_type: LeapDataType,
-                                 heatmap_function: Optional[Callable[..., npt.NDArray[np.float32]]] = None):
+                                 heatmap_function: Optional[Callable[..., npt.NDArray[np.float32]]] = None,
+                                 connects_to=None):
     def decorating_function(user_function: VisualizerCallableInterface):
         for viz_handler in leap_binder.setup_container.visualizers:
             if viz_handler.visualizer_handler_data.name == name:
@@ -125,6 +142,10 @@ def tensorleap_custom_visualizer(name: str, visualizer_type: LeapDataType,
 
         leap_binder.set_visualizer(user_function, name, visualizer_type, heatmap_function)
 
+        if connects_to is not None:
+            arg_names = leap_binder.setup_container.visualizers[-1].visualizer_handler_data.arg_names
+            _add_mapping_connections(connects_to, arg_names, NodeMappingType.Visualizer, name)
+
         def _validate_input_args(*args, **kwargs):
             for i, arg in enumerate(args):
                 assert isinstance(arg, (np.ndarray, SamplePreprocessResponse)), (
@@ -249,6 +270,64 @@ def tensorleap_preprocess():
 
     return decorating_function
 
+def tensorleap_element_instance_preprocess(instance_mask_encoder: Callable[[str, PreprocessResponse], List[ElementInstance]]):
+    def decorating_function(user_function: Callable[[], List[PreprocessResponse]]):
+        def user_function_instance() -> List[PreprocessResponse]:
+            result = user_function()
+            for preprocess_response in result:
+                sample_ids_to_instance_mappings = {}
+                instance_to_sample_ids_mappings = {}
+                instance_ids_to_names = {}
+                all_sample_ids = preprocess_response.sample_ids.copy()
+                for sample_id in preprocess_response.sample_ids:
+                    instances_masks = instance_mask_encoder(sample_id, preprocess_response)
+                    instances_ids = [f'{sample_id}_{instance_id}' for instance_id in range(len(instances_masks))]
+                    sample_ids_to_instance_mappings[sample_id] = instances_ids
+                    instance_to_sample_ids_mappings[sample_id] = sample_id
+                    instance_names = [instance.name for instance in instances_masks]
+                    instance_ids_to_names[sample_id] = 'none'
+                    for instance_id, instance_name in zip(instances_ids, instance_names):
+                        instance_to_sample_ids_mappings[instance_id] = sample_id
+                        instance_ids_to_names[instance_id] = instance_name
+                    all_sample_ids.extend(instances_ids)
+                preprocess_response.sample_ids_to_instance_mappings = sample_ids_to_instance_mappings
+                preprocess_response.instance_to_sample_ids_mappings = instance_to_sample_ids_mappings
+                preprocess_response.instance_ids_to_names = instance_ids_to_names
+                preprocess_response.sample_ids = all_sample_ids
+            return result
+
+        def builtin_instance_metadata(idx: str, preprocess: PreprocessResponse) -> Dict[str, str]:
+            return {'is_instance': '0', 'original_sample_id': idx, 'instance_name': 'none'}
+        leap_binder.set_preprocess(user_function_instance)
+        leap_binder.set_metadata(builtin_instance_metadata, "builtin_instance_metadata")
+
+        def _validate_input_args(*args, **kwargs):
+            assert len(args) == 0 and len(kwargs) == 0, \
+                (f'tensorleap_element_instance_preprocess validation failed: '
+                 f'The function should not take any arguments. Got {args} and {kwargs}.')
+
+        def _validate_result(result):
+            assert isinstance(result, list), \
+                (f'tensorleap_element_instance_preprocess validation failed: '
+                 f'The return type should be a list. Got {type(result)}.')
+            for i, response in enumerate(result):
+                assert isinstance(response, PreprocessResponse), \
+                    (f'tensorleap_element_instance_preprocess validation failed: '
+                     f'Element #{i} in the return list should be a PreprocessResponse. Got {type(response)}.')
+            assert len(set(result)) == len(result), \
+                (f'tensorleap_element_instance_preprocess validation failed: '
+                 f'The return list should not contain duplicate PreprocessResponse objects.')
+
+        def inner(*args, **kwargs):
+            _validate_input_args(*args, **kwargs)
+            result = user_function_instance()
+            _validate_result(result)
+            return result
+
+        return inner
+
+    return decorating_function
+
 
 def tensorleap_unlabeled_preprocess():
     def decorating_function(user_function: Callable[[], PreprocessResponse]):
@@ -275,7 +354,39 @@ def tensorleap_unlabeled_preprocess():
     return decorating_function
 
 
-def tensorleap_input_encoder(name: str, channel_dim=-1):
+def tensorleap_instances_masks_encoder(name: str):
+    def decorating_function(user_function: InstanceCallableInterface):
+        leap_binder.set_instance_masks(user_function, name)
+
+        def _validate_input_args(sample_id: str, preprocess_response: PreprocessResponse):
+            assert isinstance(sample_id, str), \
+                (f'tensorleap_instances_masks_encoder validation failed: '
+                 f'Argument sample_id should be str. Got {type(sample_id)}.')
+            assert isinstance(preprocess_response, PreprocessResponse), \
+                (f'tensorleap_instances_masks_encoder validation failed: '
+                 f'Argument preprocess_response should be a PreprocessResponse. Got {type(preprocess_response)}.')
+            assert type(sample_id) == preprocess_response.sample_id_type, \
+                (f'tensorleap_instances_masks_encoder validation failed: '
+                 f'Argument sample_id should be as the same type as defined in the preprocess response '
+                 f'{preprocess_response.sample_id_type}. Got {type(sample_id)}.')
+
+        def _validate_result(result):
+            assert isinstance(result, list), \
+                (f'tensorleap_instances_masks_encoder validation failed: '
+                 f'Unsupported return type. Should be a numpy array. Got {type(result)}.')
+
+        def inner(sample_id, preprocess_response):
+            _validate_input_args(sample_id, preprocess_response)
+            result = user_function(sample_id, preprocess_response)
+            _validate_result(result)
+            return result
+
+        return inner
+
+    return decorating_function
+
+
+def tensorleap_input_encoder(name: str, channel_dim=-1, model_input_index=None):
     def decorating_function(user_function: SectionCallableInterface):
         for input_handler in leap_binder.setup_container.inputs:
             if input_handler.name == name:
@@ -314,6 +425,11 @@ def tensorleap_input_encoder(name: str, channel_dim=-1):
             _validate_result(result)
             return result
 
+        node_mapping_type = NodeMappingType.Input
+        if model_input_index is not None:
+            node_mapping_type = NodeMappingType(f'Input{str(model_input_index)}')
+        inner.node_mapping = NodeMapping(name, node_mapping_type)
+
         return inner
 
     return decorating_function
@@ -354,12 +470,16 @@ def tensorleap_gt_encoder(name: str):
             _validate_result(result)
             return result
 
+        inner.node_mapping = NodeMapping(name, NodeMappingType.GroundTruth)
+
         return inner
 
+
+
     return decorating_function
 
 
-def tensorleap_custom_loss(name: str):
+def tensorleap_custom_loss(name: str, connects_to=None):
     def decorating_function(user_function: CustomCallableInterface):
         for loss_handler in leap_binder.setup_container.custom_loss_handlers:
             if loss_handler.custom_loss_handler_data.name == name:
@@ -368,6 +488,11 @@ def tensorleap_custom_loss(name: str):
 
         leap_binder.add_custom_loss(user_function, name)
 
+        if connects_to is not None:
+            arg_names = leap_binder.setup_container.custom_loss_handlers[-1].custom_loss_handler_data.arg_names
+            _add_mapping_connections(connects_to, arg_names, NodeMappingType.CustomLoss, name)
+
+
         valid_types = (np.ndarray, SamplePreprocessResponse)
         try:
             import tensorflow as tf
